(* Part of Fail-Safe C. (c) 2002-2007 Yutaka Oiwa. *)

(** {2}Additional operations on option types. *)

(** obsolete: use Option.get *)
let someof = function
    Some s -> s
  | None -> invalid_arg "Util.someof"

(** {2}Additional integer operations. *)

let rec gcd x y = if y = 0 then x else gcd y (x mod y)

(** {2}Additional float operations. *)

let frexp_int f = 
  (** returns integer significant and exponent of f.
     when (x, n) = frexp_int f, 
     x is an 64bit integer, n is integer, and
     f = x *. 2 ** n holds. *)
  if f = 0.0 then 0L, 0 else
  let rec iter s oexp =
    if (fst (modf s) = 0.0) then s, oexp
    else
      iter (s *. 2.0) (oexp - 1)
  in
  let x, e = frexp f in
  let x, e = iter x e in
  let x64 = Int64.of_float x in
  assert (Int64.to_float x64 = x);
  assert (ldexp x e = f);
  x64, e

(** {2}Additional list operations. *)

(** now using ExtLib *)

module ExtList_addition = struct
  (* copied from ExtList *)
  type 'a mut_list =  {
      hd: 'a; 
      mutable tl: 'a list
    }
  external inj : 'a mut_list -> 'a list = "%identity"
  let dummy_node () = { hd = Obj.magic (); tl = [] }

  let map_flatten f l = 
    let rec inner dst = function
	[] -> dst
      | h :: t -> 
	  let r = { hd = h; tl = [] } in
	  dst.tl <- inj r;
	  inner r t
    in
    let rec outer dst = function
      | [] -> ()
      | h :: t -> outer (inner dst (f h)) t
    in
    let r = dummy_node () in
    outer r l;
    r.tl

  let map_flatten_i f l = 
    let rec inner dst = function
	[] -> dst
      | h :: t -> 
	  let r = { hd = h; tl = [] } in
	  dst.tl <- inj r;
	  inner r t
    in
    let rec outer i dst = function
      | [] -> ()
      | h :: t -> outer (i + 1) (inner dst (f i h)) t
    in
    let r = dummy_node () in
    outer 0 r l;
    r.tl
end

let map_flatten = ExtList_addition.map_flatten
let map_flatten_i = ExtList_addition.map_flatten_i
let list_iteri = ExtList.List.iteri
let split_at_nth = ExtList.List.split_nth
let list_map = ExtList.List.map
let list_map2 = ExtList.List.map2
let list_mapi = ExtList.List.mapi
let list_append = ExtList.List.append

let list_repeat = 
  let rec iter acc e = function
    0 -> acc
  | n -> iter (e::acc) e (n - 1)
  in fun e -> iter [] e

let rec string_of_list f sep l =
    String.concat sep (List.map f l)
(*  [] -> ""
  | [e] -> f e
  | h::t -> f h ^ sep ^ string_of_list f sep t*)

(** {2}extensive array *)

module Earray = struct
  type 'a t = {
      mutable size : int; (* max. index *)
      mutable data : 'a option array;
      zero : 'a option
    }

  let empty_with_default ~zero = { size = 0; data = [| |]; zero = Some zero }
  let empty () = { size = 0; data = [| |]; zero = None }

  let copy a = 
    { a with data = Array.copy a.data }

  let mem a i = 
    if i < a.size then
      a.data.(i) <> None
    else
      false

  let get_option a i =
    if i < a.size then
      a.data.(i)
    else 
      a.zero

  let get a i = 
    match get_option a i with
      Some n -> n
    | None -> raise Not_found

  let unset a i =
    if mem a i then
      a.data.(i) <- a.zero

  let set a i v =
    let old_size = Array.length a.data in
    if i >= old_size then begin
      let new_size = old_size * 2 in
      let new_size = if i >= new_size then i + 1 else new_size in
      let new_array = 
	Array.create new_size a.zero in
      Array.blit a.data 0 new_array 0 old_size;
      a.data <- new_array
    end;
    if i >= a.size then a.size <- i + 1;
    a.data.(i) <- Some v

  let incr a i = 
    set a i (get a i + 1)

  let length a = a.size

  let iteri f a =
    for i = 0 to a.size - 1 do
      match a.data.(i) with
	None -> ()
      |	Some x -> f i x
    done

  let map f a = 
    let b = 
      match a.zero with
	Some x -> empty_with_default (f x)
      |	None -> empty ()
    in
    iteri (fun i x -> set b i (f x)) a;
    b

  let mapi f a = 
    let b = empty () in
    iteri (fun i x -> set b i (f i x)) a;
    b

  let filter_i f a = 
    let b = empty ()
    in
    iteri (fun i x -> if f i x then set b i x) a;
    b
end

type 'a earray = 'a Earray.t
let earray_get = Earray.get
let earray_set = Earray.set
let earray_length = Earray.length
let earray_iteri = Earray.iteri
let earray_mem = Earray.mem

(** {2} Other Functions. *)

let unwind_protect p b f = 
  let p = p () in
  try
    let r = b p in
    f p; r
  with e ->
    f p; raise e

let failwith_p fmt = 
  let b = Buffer.create 80 in
  Format.kfprintf 
    (fun _f ->
      let s = Buffer.contents b in
      Buffer.clear b;
      failwith s)
    (Format.formatter_of_buffer b)
    fmt
