open C_abstree
open Ctt_abstree
open C_typing2
open Locterm
open Big_int_infix
open Wl

type global_declarations = {
  global_typedefs : (identifier * c_type) list;
}

let filter_typedefs ~env =
  let typedefs = Hashtbl.fold
                   (fun k v acc -> match v.gbind_type with
                        EnumVal _ | Var _ -> acc
                      | TypeDefName tp -> (k, tp) :: acc)
                   env.global_binding [] in
  { global_typedefs = typedefs }

let rec specs_to_type ~env ~no_structdef specs =
  let is_const = ref false in
  let is_volatile = ref false in
  let extension = ref [] in
  let signedness = ref None in
  let storageclass = ref None in
  let modifier = ref None in
  let builtin_basetype = ref None in
  let full_basetype = ref None in
  let is_void = ref false in
  let repl r v = 
    match !r with
      None -> r := Some v
    | Some _ -> failwith "duplicated specifier"
  in
  let set r = match !r with
    true -> failwith "parse error (specs_to_type::set)"
  | false -> r := true
  in
  List.iter 
    (function
	TypeQualifier Const -> set is_const
      |	TypeQualifier Volatile -> set is_volatile
      |	TypeSpec (PtypespecBuiltin Signed) -> repl signedness Tsigned
      |	TypeSpec (PtypespecBuiltin Unsigned) -> repl signedness Tunsigned
      |	TypeSpec (PtypespecBuiltin Void) -> set is_void
      |	TypeSpec (PtypespecBuiltin Long) ->
	  if !modifier = Some Tlong then
	    modifier := Some Tlonglong
	  else
	    repl modifier Tlong
      |	TypeSpec (PtypespecBuiltin Short) -> repl modifier Tshort
      |	TypeSpec (PtypespecBuiltin p) -> repl builtin_basetype p
      |	ExtendedDeclSpec (id, t) ->
	  extension := !extension @ [id, t];
      | StorageClass s -> repl storageclass s
      |	TypeSpec (PtypespecEnumByName ename) ->
	  repl full_basetype type_int
      |	TypeSpec (PtypespecEnumByDef (nameopt, decl)) ->
	  repl full_basetype type_int
      |	TypeSpec (PtypespecAlias ident) ->
	  repl full_basetype (make_c_type ~env (Tabstract ident))
      | TypeSpec (PtypespecStruct (is_union, Some name, _, ext, loc)) ->
	  let id = find_named_struct ~env ~is_union ~ext:[] ~loc name in
	  repl full_basetype (make_c_type ~env (Tstruct id))
      |	TypeSpec (PtypespecStruct (is_union, None, Some def, ext, loc)) ->
          let def = { str_union_p = is_union; str_size = None; 
	              str_align = None; str_fields = []; str_fields_byname = []; str_extension = [];
                      str_assignable = false;
	              str_loc = loc
	            } in
          let def = translate_struct_by_extension ~ext def in
          let name = 
            try match List.assoc "named" def.str_extension with
                  C_abstree.Estring n -> n
                | _ -> assert false
            with
              Not_found -> assert false
          in
          let id = ref None in
          Util.earray_iteri
            (fun i desc ->
              try match List.assoc "named" desc.str_extension with
                    C_abstree.Estring n -> if name = n then id := Some i
                  | _ -> ()
              with
                Not_found -> ()
            ) env.struct_table;
          (match !id  with
            Some id -> repl full_basetype (make_c_type ~env (Tstruct id))
          | _ -> assert false
          )
      |	TypeSpec (PtypespecStruct (_, None, None, _, _)) ->
	  failwith "struct declaration must have either definition or name"
    ) specs;
  if !is_void then begin
    if !builtin_basetype != None || 
       !signedness != None || !full_basetype != None
    then failwith "void should not be qualified"
    else 
      !storageclass, !extension,
      (make_c_type ~env ~const:!is_const
	 ~volatile:!is_volatile Tvoid)
  end else match !full_basetype with
    Some fb -> begin
      if !signedness != None 
      then failwith "non-builtin should not be signedness qualified"
      else if !builtin_basetype != None 
      then failwith "non-builtin should not be used with basetypes"
      else
	!storageclass, !extension,
	{ fb with
	  ct_const_p = fb.ct_const_p || !is_const;
	  ct_volatile_p = fb.ct_volatile_p || !is_volatile;
	}
    end
  | None -> begin
    let typ = 
      match !signedness, !modifier, !builtin_basetype with
	None, None, Some Char -> Tbuiltin Tchar
      | Some Tsigned, None, Some Char -> Tbuiltin Tschar
      | Some Tunsigned, None, Some Char -> Tbuiltin Tuchar
      | None, None, Some Float -> Tbuiltin Tfloat
      | None, None, Some Double -> Tbuiltin Tdouble
      | None, Some Tlong, Some Double -> Tbuiltin Tlongdouble
      | Some Tunsigned, Some Tlonglong, (None | Some Int) -> Tbuiltin Tulonglong
      | _, Some Tlonglong, (None | Some Int) -> Tbuiltin Tlonglong
      | Some Tunsigned, Some Tlong, (None | Some Int) -> Tbuiltin Tulong
      | _, Some Tlong, (None | Some Int) -> Tbuiltin Tlong
      | Some Tunsigned, Some Tshort, (None | Some Int) -> Tbuiltin Tushort
      | _, Some Tshort, (None | Some Int) -> Tbuiltin Tshort
      | Some Tunsigned, None, (None | Some Int) -> Tbuiltin Tuint
      | _, None, (None | Some Int) -> Tbuiltin Tint
      | _ -> failwith "parse error (specs_to_type::typ)"
    in
    !storageclass, !extension, (make_c_type ~env ~const:(!is_const) ~volatile:(!is_volatile) typ)
  end

and parse_argtypes_iter ~env acc = function
    [] -> List.rev acc, false
  | (PpdeclConcrete(ts,d) |
    PpdeclAbstract(ts,d)) :: tl ->
      let storageclass, ext, basetype = (specs_to_type ~env ~no_structdef:true ts) in
        (* TODO: pass this REGISTER to environment *)
      if ext <> [] then
	failwith "parse error: extension not allowed in argument type";
      if storageclass != None && storageclass <> Some C_abstree.Register then
	failwith "parse error: storage class not allowed in argument type"
      else
	let (id,ty,_) = decl_to_type ~env d basetype in
	parse_argtypes_iter ~env ((id, ty)::acc) tl
  | [PpdeclVariant] ->
      List.rev acc, true
  | PpdeclVariant :: _ ->
      raise (TypeError_local("\"...\" must be a final argument (parser error?)"))

and parse_argtypes acc ~env = function
    [] -> [], true (* K&R style any-arg function f() *)
  | [PpdeclAbstract([TypeSpec (PtypespecBuiltin Void)], PdeclAnonymous)] -> [], false 
	(* ANSI style no-arg function f(void) *)
  | l -> parse_argtypes_iter ~env acc l

and decl_to_type ~env decl basetype =
  (* ret value = ident, type, [ident, type; ...] (inner bindings) *)
  match decl with
  | PdeclAnonymous -> "", basetype, None
  | PdeclIdent x -> x, basetype, None
  | PdeclPointer(ql,decl) ->
      decl_to_type ~env decl (make_c_type_ql ~env ql (Tpointer(basetype)))
  | PdeclArray(decl, _) ->
      if is_type_functype basetype then
	raise (TypeError_local "cannot declare array of functions");
      decl_to_type ~env decl (make_c_type ~env (Tarray(basetype,None)))
  | PdeclFuncType(decl,ptypes) -> 
      let this_bindings, is_varargs = parse_argtypes [] ~env ptypes in
      let argtypes = List.map (fun (_nam,t) -> convert_funcarg_type ~env t) this_bindings in
      let ident, rettype, inner_binding = 
	decl_to_type ~env decl (make_c_type ~env (Tfunction(argtypes, is_varargs, basetype))) in
      ident, rettype, if inner_binding <> None then inner_binding else Some this_bindings
  | PdeclFuncIdent(decl,[]) -> 
      decl_to_type ~env (PdeclFuncType(decl,[])) basetype
  | PdeclFuncIdent(decl,idents) -> 
      raise (TypeError_local "Function declaration with untyped arguments not allowed here")

let rec parse_global_declarators_rev ~loc
    ((sclass, is_typedef) as sclass')
    basetype binds_rev ~env =
  function
      [] -> binds_rev
    | PinitDecl(decl,init) :: tl ->
	let id, ty, _ = decl_to_type ~env decl basetype in
	(* Format.eprintf "%s: keep newer = %b@." nid keep_newer_initialization; *)
        let _, phys_ty = lookup_var_type_from_global ~env id in
	let binds_rev = 
          if is_typedef then
            match phys_ty with
                TypeDefName pty -> Wl.WLTypedef(id, pty)::binds_rev
              | _ -> assert false
          else
            match phys_ty with
                Var pty -> Wl.WLDeclVar(id, ty, pty)::binds_rev
              | _ -> assert false
	in
	parse_global_declarators_rev ~loc sclass' basetype binds_rev ~env tl



let filter_global_var ~env decl =
  Util.map_flatten
    (fun d -> match locval d with
      PdeclVariable(specs, decls) ->
        let sclass, ext, basetype = specs_to_type ~env specs ~no_structdef:false in
	let (_,is_typedef) as sclass = match sclass, ext with
	  Some C_abstree.Static, [] -> ModuleStatic, false
	| Some C_abstree.Extern, ext -> Extern ext, false
	| Some C_abstree.Typedef, [] -> ModuleStatic, true
	| None, [] -> Global, false
	| Some s, [] -> failwith ("bad specifier: " ^ C_pp.string_of_storage_class s)
	| _, ((id,_)::_) -> failwith (id ^ " cannot be used here")
	in
	let binds = parse_global_declarators_rev ~loc:(locget d) sclass basetype [] ~env decls in
	(* Format.eprintf "  binding: %d\n" (List.length binds); *)
	binds
    | _ -> []
    ) decl

let parse_global_declarations decl =
  let env = empty_environment decl in
  try
    (* Format.eprintf "Before Parse: %d.@." (List.length decl); *)
    let t = parse_global_declarations2 [] ~env decl in
    (* Format.eprintf "After Parse: %d.@." (List.length t); *)
    let t = filter_definitions ~env t in
    let genv = filter_environment env t in
    let genv, t = Update_struct_types.reduce_program ~genv t in
    let gv1 = filter_global_var ~env decl in
    let gv2 = ref [] in
    Util.earray_iteri
      (fun i desc ->
        let tag, name = try match List.assoc "named" desc.str_extension with
              C_abstree.Estring name -> None, Some name
            | _ -> assert false
          with
            Not_found -> None, None
        in
        gv2 := Wl.WLStruct { wl_struct_tag = tag; wl_struct_name = name; wl_struct_id = i} :: !gv2
      ) genv.Ctt_abstree.struct_table;
    genv, t, !gv2 @ gv1
  with
    TypeError_typed(e, msg) ->
      let (fname, lno, _, _) = locget e in
      Format.eprintf "%s(%d): %s: %s@." fname lno 
	(Ctt_printer.strof_expr e) msg;
      flush stderr;
      failwith "type check failed (1)"
  | TypeError_untyped(e, msg) ->
      let (fname, lno, _, _) = locget e in
      Format.eprintf "%s(%d): %a: %s@." fname lno 
	C_pp.pp_print_expression e msg;
      flush stderr;
      failwith "type check failed (2)"
  | TypeError_located(loc, msg) ->
      let (fname, lno, _, _) = loc in
      Format.eprintf "%s(%d): %s@." fname lno msg;
      flush stderr;
      failwith "type check failed (3)"

