open Arg
open Fscw_tree
open Wl
open Wl_require
open Wl_generate
open Ctt_abstree

let dprintf l f = Debug.dprintf ~category:2 l f
let get_debug_flags () = Debug.get_debug_flags ~category:2

let make_cpp_cmdline sysincludedir includedirs fname = 
  let definition_endian =
    if Fsc_config.is_littleendian
    then "-D__failsafeC_little_endian"
    else "-D__failsafeC_big_endian" in
  let args =
    ["cpp"; "-nostdinc"] @
    (match sysincludedir with Some s -> ["-idirafter"; s] | None -> []) @
    (Util.map_flatten (fun d -> ["-I"; d]) includedirs) @
    ["-ansi"; "-undef"; "-D__failsafeC"; definition_endian] @
    [fname]
  in
  (String.concat " "
     (Util.map_flatten
	(fun s -> [Filename.quote s]) args))

let separate_fscw fn =
  Fscw_tree.reset fn;
  let chan = open_in fn in
  let lexbuf = Lexing.from_channel chan in
  let tokenizer = Fscw_lexer.make_tokenizer () in
  let t = Fscw_parser.parse_fscw tokenizer lexbuf in
  t

let read_chan chan =
  C_parserhelper.reset "";
  let lexbuf = Lexing.from_channel chan in
  C_parser.translation_unit C_lexer.token lexbuf

let output_string_and_pos outch (str, (file, line)) =
  Printf.fprintf outch "#line %d \"%s\"\n" line file;
  output_string outch str

let process ~infile ~outfile sysincludedir includedirs =
  let fscw = separate_fscw infile in
  (*Printf.printf "P:%s\nC:%s\nF:%s\nR:%s\nN:%s\n"
    fscw.r_prologue fscw.r_c_prologue fscw.r_fsc_prologue fscw.r_required_decls fscw.r_native_implementation;*)

  let outch = open_out outfile in
  output_string_and_pos outch fscw.r_c_prologue;
  let (_, (_, line_1)), (_, (_, line_2)) = fscw.r_c_prologue, fscw.r_fsc_prologue in
  Printf.fprintf outch "#line %d \"%s\"\n"
                       (line_2 - line_1 + 2)
                       outfile;
  let tmpfn, tmpch = Filename.open_temp_file "fscw" ".c" in
  output_string_and_pos tmpch fscw.r_fsc_prologue;
  close_out tmpch;
  let cmdline = make_cpp_cmdline sysincludedir includedirs tmpfn in

  let chan = Unix.open_process_in cmdline in
  dprintf 1 "parsing...";
  let t =
    try read_chan chan with e -> close_in chan; raise e
  in
  let r = Unix.close_process_in chan in
  Sys.remove tmpfn;
  if r <> Unix.WEXITED 0 then
    (dprintf 1 "preprocessor failed."; exit 1);

  let t = C_abstree_remove_definitions.f t in
  let genv, t, gv = C_global_only_typing.parse_global_declarations t in
  dprintf 1 "cleaning...";
  let t = Expunge_localstatic.reduce_program ~genv t in
  let genv, t = Ctt_pad_structures.translate_program genv t in
  let genv, t = Ctt_reduce_mdarray.reduce_program ~genv t in
  let genv, t = Ctt_reduce_unions.translate_program ~genv t in
  let genv, t = Ctt_add_tentative_initializer.reduce_program ~genv t in
  (*let genv, t = Ctt_remove_unuseddecls.f ~genv t in*)
  dprintf 1 "analyzing assignments...";
  let genv, t = genv, Cttm_expunge_assignment.translate_program t in
  let genv, t = genv, Separate_side_effect.translate_program_to_il0 ~genv t in
  let genv, t = Il0_translate_bitfield.translate_program ~genv t in
  let genv, t = Il0_type_canonify.translate_program ~genv t in

  let genv2 = make_genv2 genv gv in

  let req_list_default = [
    WLR_function (WLTbuiltin Tchar);
    WLR_function (WLTbuiltin Tschar);
    WLR_function (WLTbuiltin Tuchar);
    WLR_function (WLTbuiltin Tshort);
    WLR_function (WLTbuiltin Tushort);
    WLR_function (WLTbuiltin Tint);
    WLR_function (WLTbuiltin Tuint);
    WLR_function (WLTbuiltin Tlong);
    WLR_function (WLTbuiltin Tulong);
    WLR_function (WLTbuiltin Tlonglong);
    WLR_function (WLTbuiltin Tulonglong);
    WLR_function (WLTbuiltin Tfloat);
    WLR_function (WLTbuiltin Tdouble);
    (* WLR_function (WLTbuiltin Tlongdouble); TODO *)
    WLR_function (WLTpointer);
    WLR_function (WLTstring);
    WLR_sp_unwrap (WLTstring)
  ] in
  let req_list = List.map
    (function
    | RequireTypedef name -> WLR_function (WLTabstract name)
    (*| RequireStruct name -> WLR_typedef (WLTstruct (wl_struct_of_struct_tag ~genv2 name)) *)
    | RequireStruct name -> WLR_function (WLTstruct (wl_struct_of_struct_name ~genv2 name))
    | RequireFunction name -> WLR_wrapper_decl name
    | AutoGenerate name -> WLR_autogen name
    ) (fst fscw.r_required_decls) in

  let req_list = Wl_require.resolve_required ~genv2 (req_list @ req_list_default) in

  Record_globalnames.clear ();
  List.iter (function 
             | WLR_typeinfo wty when wty <> WLTvoid ->
               let p = parse_wl_type ~genv2 wty in
               Record_globalnames.require_name (Record_globalnames.GNtypeinfo p.tp_c_t)
             | _ -> ()) req_list;
  let rr = Add_support_funcs.f_standalone ~genv [] in
  C_pp.pp_print_program (Format.formatter_of_out_channel outch) rr;

  List.iter (pp_wlr (Format.formatter_of_out_channel outch) ~genv2) req_list;

  output_string_and_pos outch fscw.r_native_implementation;
  close_out outch;

  exit 0

let _ =
  let sysincludedir = ref None in
  let includedirs = ref [] in
  let infile = ref None in
  let outfile = ref None in
  let setfile var s = var := Some s in
  if not !Sys.interactive then begin
    let argspec = Arg.align
	([
	 "-I", String (fun s -> includedirs := !includedirs @ [s]), " specify include directories";
         (*"--debug", String (Debug.parse_debug_flags), " set compiler-debugging options"; *)
         "---sysinclude",  String (setfile sysincludedir), " specify system include directory";
         "---input-file",  String (setfile infile), " specify name for .fscw input file";
         "---output-file", String (setfile outfile), " specify name for .c output file";
	 ])
    in
    let usagemsg = "Usage: fscw [options]" in
    let usage s = Arg.usage argspec (s ^ usagemsg); exit 1 in
    Arg.parse argspec (fun s -> raise (Bad "bad usage")) usagemsg;
    if !infile = None  then usage "no input file specified: ";
    if !outfile = None then usage "no output file specified: ";
    let infile  = Option.get !infile in
    let outfile = Option.get !outfile in
    Install_fsc_extension.install ();
    C_typing2.translate_struct_by_extension_hook := !C_typing.translate_struct_by_extension_hook; 
    C_typing2.translate_external_extension_hook := !C_typing.translate_external_extension_hook;
    C_typing2.fill_zero_initializer := !C_typing.fill_zero_initializer;
    C_typing2.allow_calling_undefined_functions := !C_typing.allow_calling_undefined_functions;
    process ~infile:infile ~outfile:outfile !sysincludedir !includedirs
  end

