%{(* 
   Part of Fail-Safe C Compiler. Produced by Yutaka Oiwa.
   (c) 2001-2005 Yutaka Oiwa.
   (c) 2005-     AIST.

   This file is written by Yutaka Oiwa in 2001 -- 2006. *)

(** The parser for C programs. *)

open Locterm
open C_abstree
open C_parserhelper

let parameter_scope = ref None
let location () = 
  let bpos = Parsing.symbol_start () in
  lookup_location bpos

let make_exp t = locput ~loc:(location ()) t
let make_stmt t = locput ~loc:(location ()) t
let make_decl t = locput ~loc:(location ()) t
let typedef_flag = ref false
let is_typedef () = !typedef_flag
let reset_typedef_flag () = typedef_flag := false
let set_typedef_flag () = typedef_flag := true
let typedef_stack = ref []

let debug_with_location s = 
  let (fname, lineno, _, _) = location () in
  Printf.eprintf "%s(%d): %s" fname lineno s;
  prerr_newline ()

let failwith_location s = 
  let (fname, lineno, _, _) = location () in
  failwith 
    (Printf.sprintf "%s(%d): %s" fname lineno s)

let enter_scope () =
(*debug_with_location "enter";*)
  enter_identifier_scope ();
  push typedef_stack !typedef_flag;
  typedef_flag := false

let leave_scope () = 
(*debug_with_location "leave";*)
  leave_identifier_scope ();
  typedef_flag := pop typedef_stack

let enter_merge_parameter_scope p =
(*debug_with_location "enter_merge";*)
  enter_merge_identifier_scope p;
  push typedef_stack !typedef_flag;
  typedef_flag := false

let leave_and_save_scope p = 
(*debug_with_location "leave_save";*)
  leave_and_save_identifier_scope p;
  typedef_flag := pop typedef_stack

let clear_parameter_binding () = 
(*debug_with_location "clear";*)
  parameter_scope := None

let rec get_name = function
    PdeclAnonymous -> assert false
  | PdeclIdent x -> x
  | PdeclPointer(_,x) -> get_name x
  | PdeclArray(x,_) -> get_name x
  | PdeclFuncType(x,_) -> get_name x
  | PdeclFuncIdent(x,_) -> get_name x

let parse_error _ =
  failwith_location "parse error"

%}

%token COMMA SEMICOLON COLON DOT
%token LBRACE RBRACE LBRACKET RBRACKET LPAREN RPAREN
%token AUTO REGISTER STATIC EXTERN TYPEDEF
%token VOID CHAR SHORT INT LONG FLOAT DOUBLE SIGNED UNSIGNED
%token CONST VOLATILE INLINE
%token STRUCT UNION ENUM
%token DOTDOTDOT
%token CASE DEFAULT
%token IF ELSE SWITCH
%token WHILE DO FOR
%token GOTO CONTINUE BREAK RETURN
%token SIZEOF
%token EQUAL STAREQUAL SLASHEQUAL PERCENTEQUAL PLUSEQUAL MINUSEQUAL
%token LSHIFTEQUAL RSHIFTEQUAL AMPERSANDEQUAL HATEQUAL VBAREQUAL
%token QUESTION
%token VBARVBAR AMPERSANDAMPERSAND VBAR HAT AMPERSAND
%token EQUALEQUAL EXCLEQUAL
%token LANGLE RANGLE LANGLEEQUAL RANGLEEQUAL
%token LSHIFT RSHIFT PLUS MINUS STAR SLASH PERCENT
%token TILDE EXCL
%token PLUSPLUS MINUSMINUS RARROW
%token <string> INTEGER_CONSTANT
%token <string> CHARACTER_CONSTANT
%token <string> FLOATING_CONSTANT
%token <string> STRING
%token <C_abstree.identifier> IDENTIFIER
/* %token <C_abstree.identifier> ENUMERATION_CONSTANT */
%token <C_abstree.identifier> TYPEDEF_NAME
%token <string> PRAGMA

%token EOF

%type <C_abstree.program> translation_unit
%start translation_unit

/* Fail-Safe C extensions */
%token TYPEOF

/* Hook to general extension */
%token <C_abstree.identifier> EXTENSION_INTRO

%%
translation_unit:
  initialize translation_unit_impl EOF { List.rev $2 }

initialize: 
  { clear_parameter_binding (); clear_typedef_name (); reset_typedef_flag (); }

enter_scope:
  { enter_scope () }

clear_parameter_binding: { clear_parameter_binding () }

enter_merge_parameter_scope:
  { 
    if !parameter_scope = None then
      failwith_location "parameter declaration missing"
    else
      enter_merge_parameter_scope parameter_scope
  }

translation_unit_impl:
| translation_unit_impl external_declaration	{ $2 :: $1 }
| /* empty */ { [] }

external_declaration:
| declaration clear_parameter_binding semis_opt { $1 }
| function_definition semis_opt { $1 }
| PRAGMA { make_decl(PdeclPragmatic $1) }

semis_opt:
  { () }
| SEMICOLON semis_opt { () }

/*
function_definition:
  declaration_specifiers declarator declaration_list_opt 
    function_body
    { 
      if is_typedef () then
	failwith_location "typedef with function definition"
      else begin 
	reset_typedef_flag (); 
	make_decl(PdeclFunction($1,$2,$3,$4)) 
      end
    }
|  declarator declaration_list_opt / * function name must not be typedef'ed * /
    function_body
    { make_decl(PdeclFunction([],$1,$2,$3)) }
*/

simple_declaration_start:
  storage_class_specifier simple_declaration_internal { (StorageClass($1) :: fst $2), snd $2 }
| type_qualifier simple_declaration_internal { (TypeQualifier($1) :: fst $2), snd $2 }
| type_specifier simple_declaration_forced { (TypeSpec($1) :: fst $2), snd $2 }
| extended_declaration_specifier simple_declaration_internal { ($1 :: fst $2), snd $2 }

simple_declaration_internal:
  storage_class_specifier simple_declaration_internal { (StorageClass($1) :: fst $2), snd $2 }
| type_qualifier simple_declaration_internal { (TypeQualifier($1) :: fst $2), snd $2 }
| type_specifier simple_declaration_forced { (TypeSpec($1) :: fst $2), snd $2 }
| extended_declaration_specifier simple_declaration_internal { ($1 :: fst $2), snd $2 }
| declarator { [], $1 }

simple_declaration_forced:
  storage_class_specifier simple_declaration_forced { (StorageClass($1) :: fst $2), snd $2 }
| type_qualifier simple_declaration_forced { (TypeQualifier($1) :: fst $2), snd $2 }
| type_specifier_forced simple_declaration_forced  { (TypeSpec($1) :: fst $2), snd $2 }
| extended_declaration_specifier simple_declaration_forced { ($1 :: fst $2), snd $2 }
| declarator_forced { [], $1 }

function_definition:
  simple_declaration_start declaration_list_opt function_body
    { 
      if is_typedef () then
	failwith_location "typedef with function definition"
      else begin 
	reset_typedef_flag (); 
	make_decl(PdeclFunction(fst $1,snd $1, $2, $3))
      end
    }
|  declarator declaration_list_opt /* function name must not be typedef'ed */
      function_body
    { make_decl(PdeclFunction([],$1,$2,$3)) }

/* declaration:
  declaration_specifiers init_declarator_list_opt SEMICOLON
  { reset_typedef_flag (); make_decl(PdeclVariable($1,$2)) }
*/

declaration:
  declaration_start SEMICOLON { reset_typedef_flag (); 
				make_decl(PdeclVariable(fst $1,snd $1)) }

declaration_start:
  storage_class_specifier declaration_internal { (StorageClass($1) :: fst $2), snd $2 }
| type_qualifier declaration_internal { (TypeQualifier($1) :: fst $2), snd $2 }
| type_specifier declaration_forced { (TypeSpec($1) :: fst $2), snd $2 }
| extended_declaration_specifier declaration_internal { ($1 :: fst $2), snd $2 }

declaration_internal:
  storage_class_specifier declaration_internal { (StorageClass($1) :: fst $2), snd $2 }
| type_qualifier declaration_internal { (TypeQualifier($1) :: fst $2), snd $2 }
| type_specifier declaration_forced { (TypeSpec($1) :: fst $2), snd $2 }
| extended_declaration_specifier declaration_internal { ($1 :: fst $2), snd $2 }
| init_declarator_list_opt { [], $1 }

declaration_forced:
  storage_class_specifier declaration_forced { (StorageClass($1) :: fst $2), snd $2 }
| type_qualifier declaration_forced { (TypeQualifier($1) :: fst $2), snd $2 }
| extended_declaration_specifier declaration_forced { ($1 :: fst $2), snd $2 }
| type_specifier_forced declaration_forced  { (TypeSpec($1) :: fst $2), snd $2 }
| init_declarator_forced_list_opt { [], $1 }

declaration_list_opt:
| /* opt */	{ [] }
| declaration_list	{ $1 }

declaration_list:
  declaration	{ [$1] }
| declaration_list declaration	{ $1 @ [$2] }

/*
declaration_specifiers_opt:
| declaration_specifiers	{ $1 }
| / * opt * /	{ [] }

declaration_specifiers:
  storage_class_specifier declaration_specifiers_opt
    { StorageClass($1) :: $2 }
| type_specifier_x3 declaration_specifiers_opt
    { TypeSpec($1) :: $2 }
| type_qualifier declaration_specifiers_opt
    { TypeQualifier($1) :: $2 }
| extended_declaration_specifier declaration_specifiers_opt
    { $1 :: $2 }

*/

storage_class_specifier:
  AUTO	{ Auto }
| REGISTER	{ Register }
| STATIC	{ Static }
| EXTERN	{ Extern }
| TYPEDEF	{ set_typedef_flag (); Typedef }

type_specifier_forced:
  VOID	{ PtypespecBuiltin Void }
| CHAR	{ PtypespecBuiltin Char }
| SHORT	{ PtypespecBuiltin Short }
| INT	{ PtypespecBuiltin Int }
| LONG	{ PtypespecBuiltin Long }
| FLOAT	{ PtypespecBuiltin Float }
| DOUBLE	{ PtypespecBuiltin Double }
| SIGNED	{ PtypespecBuiltin Signed }
| UNSIGNED	{ PtypespecBuiltin Unsigned }
| struct_or_union_specifier	{ $1 }
| enum_specifier	{ $1 }

type_specifier:
  type_specifier_forced { $1 }
| TYPEDEF_NAME { PtypespecAlias $1 }

type_qualifier:
  CONST		{ Const }
| VOLATILE	{ Volatile }

struct_or_union_specifier:
  struct_or_union extended_declaration_specifiers_opt
    identifier_forced LBRACE enter_scope struct_declaration_list RBRACE
    {
     leave_scope ();
     PtypespecStruct(fst $1,Some $3,Some $6,$2,snd $1)
   }
| struct_or_union extended_declaration_specifiers_opt
    LBRACE enter_scope struct_declaration_list RBRACE
    {
     leave_scope ();
     PtypespecStruct(fst $1,None,Some $5,$2,snd $1)
   }
| struct_or_union extended_declaration_specifiers_opt identifier_forced
    { 
      PtypespecStruct(fst $1,Some $3,None,$2, snd $1)
    }

struct_or_union:
  STRUCT { Struct, location() }
| UNION	{ Union, location() }

struct_declaration_list:
  struct_declaration	{ [$1] }
| struct_declaration_list struct_declaration	{ $1 @ [$2] }

init_declarator_list_opt:
| init_declarator_list	{ $1 }
| /* opt */	{ [] }

init_declarator_forced_list_opt:
| init_declarator_forced_list	{ $1 }
| /* opt */	{ [] }

init_declarator_list:
  init_declarator	{ [$1] }
| init_declarator COMMA init_declarator_forced_list
    { $1 :: $3 }

init_declarator_forced_list:
  init_declarator_forced	{ [$1] }
| init_declarator_forced COMMA init_declarator_forced_list
    { $1 :: $3 }

init_declarator:
  declarator_internal {
    match $1 with
      dcl, init_opt, None ->
	let typedef_identifier = get_name dcl in begin
	if is_typedef () then
	  if init_opt != None then
	    failwith_location "typedef with initializer"
	  else
	    add_name typedef_identifier true
	else
	  add_name typedef_identifier false
	end;
	PinitDecl(dcl, init_opt)
    | _ ->
	failwith_location "bitfield not allowed here"
  }

init_declarator_forced:
  declarator_forced_internal {
    match $1 with
      dcl, init_opt, None ->
	let typedef_identifier = get_name dcl in begin
	if is_typedef () then
	  if init_opt != None then
	    failwith_location "typedef with initializer"
	  else
	    add_name typedef_identifier true
	else
	  add_name typedef_identifier false
	end;
	PinitDecl(dcl, init_opt)
    | _ ->
	failwith_location "bitfield not allowed here"
  }

specifier_qualifier:
  type_specifier { TypeSpec $1 }
| type_qualifier { TypeQualifier $1 }

specifier_qualifier_list:
  specifier_qualifier
    { [$1] }
| specifier_qualifier specifier_qualifier_list
    { $1 :: $2 }

struct_declaration:
  struct_declaration_start SEMICOLON {
    PstructDecl(fst $1, snd $1);
  }

struct_declaration_start:
  type_specifier struct_declaration_internal_forced { (TypeSpec $1 :: fst $2), snd $2 }
| type_qualifier struct_declaration_internal { (TypeQualifier $1 :: fst $2), snd $2 }

struct_declaration_internal:
  type_specifier struct_declaration_internal_forced { (TypeSpec $1 :: fst $2), snd $2 }
| type_qualifier struct_declaration_internal { (TypeQualifier $1 :: fst $2), snd $2 }
| struct_declarator_list { [], $1 }

struct_declaration_internal_forced:
  type_specifier_forced struct_declaration_internal_forced { (TypeSpec $1 :: fst $2), snd $2 }
| type_qualifier struct_declaration_internal_forced { (TypeQualifier $1 :: fst $2), snd $2 }
| struct_declarator_list_forced { [], $1 }

struct_declarator_list:
  struct_declarator	{ [$1] }
| struct_declarator COMMA struct_declarator_list_forced
    { $1 :: $3 }

struct_declarator_list_forced:
  struct_declarator_forced	{ [$1] }
| struct_declarator_forced COMMA struct_declarator_list_forced
    { $1 :: $3 }


struct_declarator:
  declarator_internal 
    {
     match $1 with 
       dcl, None, None -> PstructDeclNormal dcl
     | dcl, None, Some wid -> PstructDeclBitfield(Some dcl,wid)
     | _ -> failwith_location "struct/union element with initializer not allowed"
   } 
| COLON constant_expression
    { PstructDeclBitfield(None,$2) }

struct_declarator_forced:
  declarator_forced_internal 
    {
     match $1 with 
       dcl, None, None -> PstructDeclNormal dcl
     | dcl, None, Some wid -> PstructDeclBitfield(Some dcl,wid)
     | _ -> failwith_location "struct/union element with initializer not allowed"
   } 
| COLON constant_expression
    { PstructDeclBitfield(None,$2) }

enum_specifier:
  ENUM identifier_forced LBRACE enumerator_list RBRACE
    { PtypespecEnumByDef(Some $2,$4) }
| ENUM LBRACE enumerator_list RBRACE
    { PtypespecEnumByDef(None,$3) }
| ENUM identifier_forced { PtypespecEnumByName $2 }

enumerator_list:
  enumerator	{ [$1] }
| enumerator_list COMMA enumerator	{ $1 @ [$3] }

enumerator:
  identifier_forced	{ $1, None }
| identifier_forced EQUAL constant_expression	{ $1, Some $3 }

declarator:
  declarator_internal {
    match $1 with
      dcl, None, None ->
	let typedef_identifier = get_name dcl in begin
	  if is_typedef () then
	    failwith_location ("typedef with parameter or function definition : " ^ typedef_identifier)
	  else
	    add_name typedef_identifier false
	end;
	dcl
    | x, Some _, _ -> failwith_location "initializer not allowed here"
    | x, None, Some _ -> failwith_location "bitfield not allowed here"
  } 

declarator_internal:
  declarator_internal_2 { $1, None, None }
| declarator_internal_2 EQUAL c_initializer { $1, Some $3, None }
| declarator_internal_2 COLON constant_expression { $1, None, Some $3 }

declarator_internal_2:
  direct_declarator { $1 }
| STAR type_qualifier_list_opt declarator_forced_internal_2 { PdeclPointer($2,$3) }

declarator_forced:
  declarator_forced_internal {
    match $1 with
      dcl, None, None ->
	let typedef_identifier = get_name dcl in begin
	  if is_typedef () then
	    failwith_location ("typedef with parameter or function definition : " ^ typedef_identifier)
	  else
	    add_name typedef_identifier false
	end;
	dcl
    | x, Some _, _ -> failwith_location "initializer not allowed here"
    | x, None, Some _ -> failwith_location "bitfield not allowed here"
  }

declarator_forced_internal:
  declarator_forced_internal_2 { $1, None, None }
| declarator_forced_internal_2 EQUAL c_initializer { $1, Some $3, None }
| declarator_forced_internal_2 COLON constant_expression { $1, None, Some $3 }

declarator_forced_internal_2:
  direct_declarator_forced { $1 }
| STAR type_qualifier_list_opt declarator_forced_internal_2 { PdeclPointer($2,$3) }

direct_declarator:
  identifier	{ PdeclIdent($1) }
| LPAREN declarator_internal_2 RPAREN	{ $2 }
| direct_declarator LBRACKET constant_expression_opt RBRACKET
    { PdeclArray($1,$3) }
| direct_declarator LPAREN parameter_type_list RPAREN
    { PdeclFuncType($1,$3) }
| direct_declarator LPAREN parameter_identifier_list_opt RPAREN
    { PdeclFuncIdent($1,$3) }

direct_declarator_forced:
| identifier_forced	{ PdeclIdent($1) }
| LPAREN declarator_forced_internal_2 RPAREN	{ $2 }
| direct_declarator_forced LBRACKET constant_expression_opt RBRACKET
    { PdeclArray($1,$3) }
| direct_declarator_forced LPAREN parameter_type_list RPAREN
    { PdeclFuncType($1,$3) }
| direct_declarator_forced LPAREN parameter_identifier_list_opt RPAREN
    { PdeclFuncIdent($1,$3) }

type_qualifier_list_opt:
  /* opt */	{ [] }
| type_qualifier_list	{ $1 }

type_qualifier_list:
  type_qualifier	{ [$1] }
| type_qualifier_list type_qualifier	{ $1 @ [$2] }

parameter_type_list_opt:
  /* opt */	{ [] }
| parameter_type_list	{ $1 }

parameter_type_list:
  enter_scope parameter_list
    { 
      leave_and_save_scope parameter_scope; 
      $2
    }
| enter_scope parameter_list COMMA DOTDOTDOT
    {
     leave_and_save_scope parameter_scope;
     $2 @ [PpdeclVariant]
   }

parameter_list:
  parameter_declaration	{ [$1] }
| parameter_list COMMA parameter_declaration	{ $1 @ [$3] }

parameter_declaration:
  parameter_declaration_start {
    if fst (snd $1) then
      PpdeclConcrete(fst $1, snd (snd $1))
    else
      PpdeclAbstract(fst $1, snd (snd $1))
  } 

parameter_declaration_start:
  storage_class_specifier parameter_declaration_internal { 
    match $1 with
      Register | Auto -> (StorageClass($1):: fst $2), snd $2 
    | Inline | Static | Extern | Typedef ->
	failwith_location "No storage class nor typedef allowed in parameter list" 
  }
| type_qualifier parameter_declaration_internal { (TypeQualifier($1) :: fst $2), snd $2 }
| type_specifier parameter_declaration_internal_forced { (TypeSpec($1) :: fst $2), snd $2 }

parameter_declaration_internal:
  storage_class_specifier parameter_declaration_internal { 
    match $1 with
      Register | Auto -> (StorageClass($1):: fst $2), snd $2 
    | Inline | Static | Extern | Typedef
      -> failwith_location "No storage class nor typedef allowed in parameter list" 
  }
| type_qualifier parameter_declaration_internal { (TypeQualifier($1) :: fst $2), snd $2 }
| type_specifier parameter_declaration_internal_forced { (TypeSpec($1) :: fst $2), snd $2 }
| declarator
    { [], (true, $1) }
| abstract_declarator_opt { [], (false, $1) }

parameter_declaration_internal_forced:
  storage_class_specifier parameter_declaration_internal_forced {
    match $1 with
      Register | Auto -> (StorageClass($1):: fst $2), snd $2 
    | Inline | Static | Extern | Typedef
      -> failwith_location "No storage class nor typedef allowed in parameter list" 
    }
| type_qualifier parameter_declaration_internal_forced { (TypeQualifier($1) :: fst $2), snd $2 }
| type_specifier_forced parameter_declaration_internal_forced { (TypeSpec($1) :: fst $2), snd $2 }
| declarator_forced
    { [], (true, $1) }
| abstract_declarator_opt { [], (false, $1) }

/* not used
identifier_list_opt:
  / * opt * /	{ [] }
| identifier_list	{ $1 }
*/

identifier_list:
  identifier	{ [$1] }
| identifier_list COMMA identifier
    { $1 @ [$3] }

parameter_identifier_list_opt:
  /* opt */ { enter_scope (); leave_and_save_scope parameter_scope; [] }
| enter_scope parameter_identifier_list	{ leave_and_save_scope parameter_scope; $2 }

parameter_identifier_list:
  identifier	{ add_name $1 false; [$1] }
| identifier COMMA identifier_list
    { add_name $1 false; $1 :: $3 }

/* "initializer" conflicts with OCaml keyword, renamed to c_initializer */
c_initializer:
  assignment_expression	{ PinitExp($1) }
| LBRACE initializer_list_rev RBRACE	{ PinitList(List.rev $2) }
| LBRACE initializer_list_rev COMMA RBRACE	{ PinitList(List.rev $2) }

initializer_list_rev:
  c_initializer	{ [$1] }
| initializer_list_rev COMMA c_initializer	{ $3 :: $1 }

type_name:
  specifier_qualifier_list abstract_declarator_opt
    { Ptypename($1,$2) }

abstract_declarator:
  direct_abstract_declarator { $1 }
| STAR type_qualifier_list_opt abstract_declarator_opt { PdeclPointer($2,$3) }

abstract_declarator_opt:
| /* opt */ { PdeclAnonymous }
| direct_abstract_declarator { $1 }
| STAR type_qualifier_list_opt abstract_declarator_opt { PdeclPointer($2,$3) }

direct_abstract_declarator:
  LPAREN abstract_declarator RPAREN	{ $2 }
| LBRACKET constant_expression_opt RBRACKET
    { PdeclArray(PdeclAnonymous,$2) }
| direct_abstract_declarator LBRACKET constant_expression_opt RBRACKET
    { PdeclArray($1,$3) }
| direct_abstract_declarator LPAREN parameter_type_list_opt RPAREN
    { PdeclFuncType($1,$3) }
| LPAREN parameter_type_list_opt RPAREN
    { PdeclFuncType(PdeclAnonymous,$2) }

statement:
  labelled_statement	{ make_stmt($1) }
| expression_statement	{ make_stmt(PstmtExpr $1) }
| compound_statement	{ $1 }
| selection_statement	{ make_stmt($1) }
| iteration_statement	{ make_stmt($1) }
| jump_statement	{ make_stmt($1) }

labelled_statement:
  identifier COLON statement	{ PstmtLabeled($1,$3) }
| CASE constant_expression COLON statement	{ PstmtCase_Labeled($2,$4) }
| DEFAULT COLON statement	{ PstmtDefault_Labeled($3) }

expression_statement:
  expression_opt SEMICOLON	{ ($1) }

compound_statement:
  LBRACE enter_scope declaration_list_opt clear_parameter_binding statement_list_opt RBRACE
    { leave_scope (); make_stmt(PstmtCompound($3,$5)) }

function_body:
  LBRACE enter_merge_parameter_scope declaration_list_opt statement_list_opt RBRACE
    { leave_scope (); make_stmt(PstmtCompound($3,$4)) }

statement_list_opt:
  /* opt */	{ [] }
| statement_list	{ $1 }

statement_list:
  statement	{ [$1] }
| statement statement_list	{ $1 :: $2 }

selection_statement:
  IF LPAREN expression RPAREN statement	{ PstmtIf($3,$5,None) }
| IF LPAREN expression RPAREN statement ELSE statement	{ PstmtIf($3,$5,Some $7) }
| SWITCH LPAREN expression RPAREN statement	{ PstmtSwitch($3,$5) }

iteration_statement:
  WHILE LPAREN expression RPAREN statement	{ PstmtWhile($3,$5) }
| DO statement WHILE LPAREN expression RPAREN SEMICOLON	{ PstmtDoWhile($2,$5) }
| FOR LPAREN expression_opt SEMICOLON expression_opt SEMICOLON expression_opt RPAREN statement
    { PstmtFor($3,$5,$7,$9) }

jump_statement:
  GOTO identifier SEMICOLON	{ PstmtGoto($2) }
| CONTINUE SEMICOLON	{ PstmtContinue }
| BREAK SEMICOLON	{ PstmtBreak }
| RETURN expression_opt SEMICOLON	{ PstmtReturn($2) }

expression_opt:
  /* opt */	{ None }
| expression	{ Some $1 }

expression:
  assignment_expression { $1 }
| expression COMMA assignment_expression
    { make_exp (PexpComma($1,$3)) }

assignment_expression:
  conditional_expression { $1 }
| unary_expression EQUAL assignment_expression
    { make_exp (PexpAssign($1,$3)) }
| unary_expression mod_assignment_operator assignment_expression
    { make_exp (PexpBinAssign($2,$1,$3)) }

mod_assignment_operator:
  STAREQUAL	{ PbinTimes }
| SLASHEQUAL	{ PbinDiv }
| PERCENTEQUAL	{ PbinModulo }
| PLUSEQUAL	{ PbinPlus }
| MINUSEQUAL	{ PbinMinus }
| LSHIFTEQUAL	{ PbinLshift }
| RSHIFTEQUAL	{ PbinRshift }
| AMPERSANDEQUAL	{ PbinIntAnd }
| HATEQUAL	{ PbinIntXor }
| VBAREQUAL	{ PbinIntOr }

conditional_expression:
  logical_or_expression { $1 }
| logical_or_expression QUESTION expression COLON conditional_expression
    { make_exp (PexpConditional($1,$3,$5)) }

constant_expression_opt:
  /* opt */	{ None }
| constant_expression		{ Some $1 }

constant_expression:
  conditional_expression  { $1 }

logical_or_expression:
  logical_and_expression  { $1 }
| logical_or_expression VBARVBAR logical_and_expression
    { make_exp (PexpBinExpr(PbinLogOr,$1,$3)) }

logical_and_expression:
  inclusive_or_expression { $1 }
| logical_and_expression AMPERSANDAMPERSAND inclusive_or_expression
    { make_exp (PexpBinExpr(PbinLogAnd,$1,$3)) }

inclusive_or_expression:
  exclusive_or_expression { $1 }
| inclusive_or_expression VBAR exclusive_or_expression
    { make_exp (PexpBinExpr(PbinIntOr,$1,$3)) }

exclusive_or_expression:
  and_expression { $1 }
| exclusive_or_expression HAT and_expression
    { make_exp (PexpBinExpr(PbinIntXor,$1,$3)) }

and_expression:
  equality_expression { $1 }
| and_expression AMPERSAND equality_expression
    { make_exp (PexpBinExpr(PbinIntAnd,$1,$3)) }

equality_expression:
  relational_expression { $1 }
| equality_expression EQUALEQUAL relational_expression
    { make_exp (PexpBinExpr(PbinEqual,$1,$3)) }
| equality_expression EXCLEQUAL relational_expression
    { make_exp (PexpBinExpr(PbinNotEqual,$1,$3)) }

relational_expression:
  shift_expression { $1 }
| relational_expression LANGLE shift_expression
    { make_exp (PexpBinExpr(PbinLessThan,$1,$3)) }
| relational_expression RANGLE shift_expression
    { make_exp (PexpBinExpr(PbinGtrThan,$1,$3)) }
| relational_expression LANGLEEQUAL shift_expression
    { make_exp (PexpBinExpr(PbinLessEqual,$1,$3)) }
| relational_expression RANGLEEQUAL shift_expression
    { make_exp (PexpBinExpr(PbinGtrEqual,$1,$3)) }

shift_expression:
  additive_expression { $1 }
| shift_expression LSHIFT additive_expression
    { make_exp (PexpBinExpr(PbinLshift,$1,$3)) }
| shift_expression RSHIFT additive_expression
    { make_exp (PexpBinExpr(PbinRshift,$1,$3)) }

additive_expression:
  multiplicative_expression { $1 }
| additive_expression PLUS multiplicative_expression
    { make_exp (PexpBinExpr(PbinPlus,$1,$3)) }
| additive_expression MINUS multiplicative_expression
    { make_exp (PexpBinExpr(PbinMinus,$1,$3)) }

multiplicative_expression:
  cast_expression { $1 }
| multiplicative_expression STAR cast_expression
    { make_exp (PexpBinExpr(PbinTimes,$1,$3)) }
| multiplicative_expression SLASH cast_expression
    { make_exp (PexpBinExpr(PbinDiv,$1,$3)) }
| multiplicative_expression PERCENT cast_expression
    { make_exp (PexpBinExpr(PbinModulo,$1,$3)) }

cast_expression:
  unary_expression { $1 }
| LPAREN type_name RPAREN cast_expression
    { make_exp (PexpCast($2,$4)) }

unary_expression:
  postfix_expression { $1 }
| PLUSPLUS unary_expression	{ make_exp (PexpPreInc $2) }
| MINUSMINUS unary_expression	{ make_exp (PexpPreDec $2) }
| AMPERSAND cast_expression { make_exp (PexpAddress $2) }
| STAR cast_expression { make_exp (PexpPtrDeref $2) }
| simple_unary_operator cast_expression { make_exp (PexpUnaryExpr($1,$2)) }
| SIZEOF unary_expression	{ make_exp (PexpSizeOfExpr($2)) }
| SIZEOF LPAREN type_name RPAREN	{ make_exp (PexpSizeOfType($3)) }
/* Fail-Safe C extensions */
| TYPEOF LPAREN type_name RPAREN	{ make_exp (PexpTypeOfType($3)) }

simple_unary_operator:
| PLUS	{ UnaryPlus }
| MINUS	{ UnaryMinus }
| TILDE	{ IntNot }
| EXCL	{ LogNot }

postfix_expression:
  primary_expression { $1 }
| postfix_expression LBRACKET expression RBRACKET
    { make_exp (PexpArrayRef($1,$3)) }
| postfix_expression LPAREN argument_expression_list_opt RPAREN
    { make_exp (PexpInvoke($1,$3)) }
| postfix_expression DOT identifier_forced	{ make_exp (PexpField($1,$3)) }
| postfix_expression RARROW identifier_forced	{ make_exp (PexpPtrField($1,$3)) }
| postfix_expression PLUSPLUS	{ make_exp (PexpPostInc $1) }
| postfix_expression MINUSMINUS	{ make_exp (PexpPostDec $1) }

primary_expression:
  identifier	{ make_exp (PexpVar $1) }
| constant	{ make_exp (PexpConstant $1) }
| string	{ make_exp (PexpConstant $1) }
| LPAREN expression RPAREN	{ $2 }

argument_expression_list_opt:
  /* opt */	{ [] }
| argument_expression_list { $1 }

argument_expression_list:
  assignment_expression	{ [$1] }
| argument_expression_list COMMA assignment_expression	{ $1 @ [$3] }

constant:
  integer_constant { $1 }
| character_constant { $1 }
| floating_constant { $1 }

/* terminals */

integer_constant:
  INTEGER_CONSTANT	{ PconstInteger $1 }

character_constant:
  CHARACTER_CONSTANT	{ PconstChar $1 }

floating_constant:
  FLOATING_CONSTANT	{ PconstFloat $1 }

identifier:
  IDENTIFIER	{ $1 }

string:
  string_internal { PconstString($1) }

string_internal:
  STRING	{ [$1] }
| STRING string_internal { $1 :: $2 }

/* enumeration_constant:
  ENUMERATION_CONSTANT	{ PconstEnum($1) } */

identifier_forced:
  IDENTIFIER	{ $1 }
| TYPEDEF_NAME	{ $1 }

/* extension hook */
extended_declaration_specifiers_opt:
  /* opt */ { [] }
| EXTENSION_INTRO extension_element extended_declaration_specifiers_opt { ($1, $2) :: $3 }

extended_declaration_specifier:
    EXTENSION_INTRO extension_element { ExtendedDeclSpec($1, $2) }

extension_tree_comma_list:
  extension_element_list { [$1] }
| extension_element_list COMMA extension_tree_comma_list { $1 :: $3 }

extension_element_list:
  extension_element_list_internal {
    match $1 with
      [i] -> i
    | l -> Elist l
}

extension_element_list_internal:
  extension_element extension_element_list_internal { $1 :: $2 }
| /* */ { [] }

extension_element:
  LPAREN extension_tree_comma_list RPAREN { Eplist $2 }
| identifier_forced { Eident $1 }
| INTEGER_CONSTANT { Eint $1 }
| STRING { Estring $1 }
